@@ -80,11 +80,11 @@ int Stream::open_file(const std::string_view &path) {
 
   std::string_view filename;
 
-  auto it = std::find(std::rbegin(path), std::rend(path), '/').base();
-  if (it == std::end(path)) {
+  auto it = std::ranges::find(std::rbegin(path), std::rend(path), '/').base();
+  if (it == std::ranges::end(path)) {
     filename = "index.html"sv;
   } else {
-    filename = std::string_view{it, static_cast<size_t>(std::end(path) - it)};
+    filename = std::string_view{it, std::ranges::end(path)};
     if (filename == ".."sv || filename == "."sv) {
       std::cerr << "Invalid file name: " << filename << std::endl;
       return -1;
@@ -398,7 +398,7 @@ int recv_version_negotiation(ngtcp2_conn *conn, const ngtcp2_pkt_hd *hd,
 
 void Client::recv_version_negotiation(const uint32_t *sv, size_t nsv) {
   offered_versions_.resize(nsv);
-  std::copy_n(sv, nsv, std::begin(offered_versions_));
+  std::ranges::copy_n(sv, nsv, std::ranges::begin(offered_versions_));
 }
 
 namespace {
@@ -556,7 +556,7 @@ int extend_max_stream_data(ngtcp2_conn *conn, int64_t stream_id,
 
 int Client::extend_max_stream_data(int64_t stream_id, uint64_t max_data) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   if (nghttp3_buf_len(&stream->reqbuf)) {
@@ -726,8 +726,8 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     settings.pmtud_probeslen = config.pmtud_probes.size();
 
     if (!config.max_udp_payload_size) {
-      settings.max_tx_udp_payload_size = *std::max_element(
-        std::begin(config.pmtud_probes), std::end(config.pmtud_probes));
+      settings.max_tx_udp_payload_size =
+        *std::ranges::max_element(config.pmtud_probes);
     }
   }
 
@@ -982,6 +982,8 @@ int Client::write_streams() {
     std::span{tx_.data.data(), std::max(ngtcp2_conn_get_send_quantum(conn_),
                                         path_max_udp_payload_size)};
   auto buf = txbuf;
+  auto pkt = std::span<const uint8_t>{};
+  auto extra_pkt = std::span<const uint8_t>{};
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
@@ -993,7 +995,7 @@ int Client::write_streams() {
     Stream *stream = nullptr;
 
     if (!sendq_.empty() && ngtcp2_conn_get_max_data_left(conn_)) {
-      stream = *std::begin(sendq_);
+      stream = *std::ranges::begin(sendq_);
 
       stream_id = stream->stream_id;
       vec.base = stream->reqbuf.pos;
@@ -1016,13 +1018,13 @@ int Client::write_streams() {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
         stream->reqbuf.pos += ndatalen;
         if (nghttp3_buf_len(&stream->reqbuf) == 0) {
-          sendq_.erase(std::begin(sendq_));
+          sendq_.erase(std::ranges::begin(sendq_));
         }
         continue;
       }
@@ -1037,34 +1039,24 @@ int Client::write_streams() {
     } else if (ndatalen >= 0) {
       stream->reqbuf.pos += ndatalen;
       if (nghttp3_buf_len(&stream->reqbuf) == 0) {
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
       }
     }
 
     if (nwrite == 0) {
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-      if (!data.empty()) {
-        auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-
-        if (auto [rest, rv] =
-              send_packet(ep, prev_ps.path.remote, prev_ecn, data, gso_size);
-            rv != NETWORK_ERR_OK) {
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, prev_ps.path.remote, prev_ecn, rest, gso_size);
-        }
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      if (pkt.empty()) {
+        return 0;
       }
 
-      // We are congestion limited.
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      break;
     }
 
-    auto last_pkt = buf.first(nwrite);
+    auto last_pkt_pos = std::ranges::begin(buf);
 
     buf = buf.subspan(nwrite);
 
-    if (std::begin(last_pkt) == std::begin(txbuf)) {
+    if (last_pkt_pos == std::ranges::begin(txbuf)) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_ecn = pi.ecn;
       gso_size = nwrite;
@@ -1072,53 +1064,46 @@ int Client::write_streams() {
                static_cast<size_t>(nwrite) > gso_size ||
                (gso_size > path_max_udp_payload_size &&
                 static_cast<size_t>(nwrite) != gso_size)) {
-      auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), std::begin(last_pkt)};
-
-      if (auto [rest, rv] =
-            send_packet(ep, prev_ps.path.remote, prev_ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, prev_ps.path.remote, prev_ecn, rest, gso_size);
-        on_send_blocked(*static_cast<Endpoint *>(ps.path.user_data),
-                        ps.path.remote, pi.ecn, last_pkt, last_pkt.size());
-      } else {
-        auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-
-        if (auto [rest, rv] = send_packet(ep, ps.path.remote, pi.ecn, last_pkt,
-                                          last_pkt.size());
-            rv != 0) {
-          assert(rest.size() == last_pkt.size());
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, ps.path.remote, pi.ecn, rest, rest.size());
-        }
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), last_pkt_pos};
+      extra_pkt = std::span{last_pkt_pos, std::ranges::begin(buf)};
+      break;
     }
 
     if (buf.size() < path_max_udp_payload_size ||
         static_cast<size_t>(nwrite) < gso_size) {
-      auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-
-      if (auto [rest, rv] =
-            send_packet(ep, ps.path.remote, pi.ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, ps.path.remote, pi.ecn, rest, gso_size);
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      break;
     }
   }
+
+  if (send_packet_or_blocked(prev_ps.path, prev_ecn, pkt, gso_size) != 0) {
+    if (!extra_pkt.empty()) {
+      on_send_blocked(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+    }
+  } else if (!extra_pkt.empty()) {
+    send_packet_or_blocked(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+  }
+
+  ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+
+  return 0;
+}
+
+int Client::send_packet_or_blocked(const ngtcp2_path &path, unsigned int ecn,
+                                   std::span<const uint8_t> data,
+                                   size_t gso_size) {
+  auto &ep = *static_cast<Endpoint *>(path.user_data);
+
+  auto [rest, rv] = send_packet(ep, path.remote, ecn, data, gso_size);
+  if (rv != 0) {
+    assert(NETWORK_ERR_SEND_BLOCKED == rv);
+
+    on_send_blocked(path, ecn, rest, gso_size);
+
+    return rv;
+  }
+
+  return 0;
 }
 
 void Client::update_timer() {
@@ -1488,8 +1473,7 @@ Client::send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
     for (; !data.empty();) {
       auto len = std::min(gso_size, data.size());
 
-      auto [_, rv] =
-        send_packet(ep, remote_addr, ecn, {std::begin(data), len}, len);
+      auto [_, rv] = send_packet(ep, remote_addr, ecn, data.first(len), len);
       if (rv != 0) {
         return {data, rv};
       }
@@ -1602,9 +1586,8 @@ Client::send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
   return {{}, NETWORK_ERR_OK};
 }
 
-void Client::on_send_blocked(const Endpoint &ep, const ngtcp2_addr &remote_addr,
-                             unsigned int ecn, std::span<const uint8_t> data,
-                             size_t gso_size) {
+void Client::on_send_blocked(const ngtcp2_path &path, unsigned int ecn,
+                             std::span<const uint8_t> data, size_t gso_size) {
   assert(tx_.num_blocked || !tx_.send_blocked);
   assert(tx_.num_blocked < 2);
   assert(gso_size);
@@ -1613,9 +1596,11 @@ void Client::on_send_blocked(const Endpoint &ep, const ngtcp2_addr &remote_addr,
 
   auto &p = tx_.blocked[tx_.num_blocked++];
 
-  memcpy(&p.remote_addr.su, remote_addr.addr, remote_addr.addrlen);
+  memcpy(&p.remote_addr.su, path.remote.addr, path.remote.addrlen);
 
-  p.remote_addr.len = remote_addr.addrlen;
+  auto &ep = *static_cast<Endpoint *>(path.user_data);
+
+  p.remote_addr.len = path.remote.addrlen;
   p.endpoint = &ep;
   p.ecn = ecn;
   p.data = data;
@@ -1703,7 +1688,7 @@ int Client::handle_error() {
 
 int Client::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   sendq_.erase(stream.get());
@@ -1787,7 +1772,7 @@ int Client::submit_http_request(Stream *stream) {
 int Client::recv_stream_data(uint32_t flags, int64_t stream_id,
                              std::span<const uint8_t> data) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   ngtcp2_conn_extend_max_stream_offset(conn_, stream_id, data.size());
@@ -1808,7 +1793,7 @@ int Client::recv_stream_data(uint32_t flags, int64_t stream_id,
 int Client::acked_stream_data_offset(int64_t stream_id, uint64_t offset,
                                      uint64_t datalen) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
   (void)stream;
   assert(static_cast<uint64_t>(stream->reqbuf.end - stream->reqbuf.begin) >=
@@ -2624,7 +2609,7 @@ int main(int argc, char **argv) {
         }
         auto l = util::split_str(optarg);
         config.available_versions.resize(l.size());
-        auto it = std::begin(config.available_versions);
+        auto it = std::ranges::begin(config.available_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
@@ -2656,7 +2641,7 @@ int main(int argc, char **argv) {
                     << max_preferred_versionslen << std::endl;
         }
         config.preferred_versions.resize(l.size());
-        auto it = std::begin(config.preferred_versions);
+        auto it = std::ranges::begin(config.preferred_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
@@ -2793,18 +2778,16 @@ int main(int argc, char **argv) {
 
   if (!ngtcp2_is_reserved_version(config.version)) {
     if (!config.preferred_versions.empty() &&
-        std::find(std::begin(config.preferred_versions),
-                  std::end(config.preferred_versions),
-                  config.version) == std::end(config.preferred_versions)) {
+        std::ranges::find(config.preferred_versions, config.version) ==
+          std::ranges::end(config.preferred_versions)) {
       std::cerr << "preferred-version: must include version " << std::hex
                 << "0x" << config.version << std::dec << std::endl;
       exit(EXIT_FAILURE);
     }
 
     if (!config.available_versions.empty() &&
-        std::find(std::begin(config.available_versions),
-                  std::end(config.available_versions),
-                  config.version) == std::end(config.available_versions)) {
+        std::ranges::find(config.available_versions, config.version) ==
+          std::ranges::end(config.available_versions)) {
       std::cerr << "available-versions: must include version " << std::hex
                 << "0x" << config.version << std::dec << std::endl;
       exit(EXIT_FAILURE);
