@@ -28,6 +28,7 @@
 #include <iostream>
 #include <fstream>
 #include <limits>
+#include <algorithm>
 
 #include <ngtcp2/ngtcp2_crypto_wolfssl.h>
 
@@ -53,8 +54,6 @@ int alpn_select_proto_h3_cb(WOLFSSL *ssl, const unsigned char **out,
   auto conn_ref =
     static_cast<ngtcp2_crypto_conn_ref *>(wolfSSL_get_app_data(ssl));
   auto h = static_cast<HandlerBase *>(conn_ref->user_data);
-  const uint8_t *alpn;
-  size_t alpnlen;
   // This should be the negotiated version, but we have not set the
   // negotiated version when this callback is called.
   auto version = ngtcp2_conn_get_client_chosen_version(h->conn());
@@ -62,8 +61,6 @@ int alpn_select_proto_h3_cb(WOLFSSL *ssl, const unsigned char **out,
   switch (version) {
   case NGTCP2_PROTO_VER_V1:
   case NGTCP2_PROTO_VER_V2:
-    alpn = H3_ALPN_V1;
-    alpnlen = str_size(H3_ALPN_V1);
     break;
   default:
     if (!config.quiet) {
@@ -73,16 +70,17 @@ int alpn_select_proto_h3_cb(WOLFSSL *ssl, const unsigned char **out,
     return SSL_TLSEXT_ERR_ALERT_FATAL;
   }
 
-  for (auto p = in, end = in + inlen; p + alpnlen <= end; p += *p + 1) {
-    if (std::equal(alpn, alpn + alpnlen, p)) {
-      *out = p + 1;
-      *outlen = *p;
+  for (auto s = std::span{in, inlen}; s.size() >= H3_ALPN_V1.size();
+       s = s.subspan(s[0] + 1)) {
+    if (std::ranges::equal(H3_ALPN_V1, s.first(H3_ALPN_V1.size()))) {
+      *out = &s[1];
+      *outlen = s[0];
       return SSL_TLSEXT_ERR_OK;
     }
   }
 
   if (!config.quiet) {
-    std::cerr << "Client did not present ALPN " << &alpn[1] << std::endl;
+    std::cerr << "Client did not present ALPN " << &H3_ALPN_V1[1] << std::endl;
   }
 
   return SSL_TLSEXT_ERR_ALERT_FATAL;
@@ -96,8 +94,6 @@ int alpn_select_proto_hq_cb(WOLFSSL *ssl, const unsigned char **out,
   auto conn_ref =
     static_cast<ngtcp2_crypto_conn_ref *>(wolfSSL_get_app_data(ssl));
   auto h = static_cast<HandlerBase *>(conn_ref->user_data);
-  const uint8_t *alpn;
-  size_t alpnlen;
   // This should be the negotiated version, but we have not set the
   // negotiated version when this callback is called.
   auto version = ngtcp2_conn_get_client_chosen_version(h->conn());
@@ -105,8 +101,6 @@ int alpn_select_proto_hq_cb(WOLFSSL *ssl, const unsigned char **out,
   switch (version) {
   case NGTCP2_PROTO_VER_V1:
   case NGTCP2_PROTO_VER_V2:
-    alpn = HQ_ALPN_V1;
-    alpnlen = str_size(HQ_ALPN_V1);
     break;
   default:
     if (!config.quiet) {
@@ -116,16 +110,17 @@ int alpn_select_proto_hq_cb(WOLFSSL *ssl, const unsigned char **out,
     return SSL_TLSEXT_ERR_ALERT_FATAL;
   }
 
-  for (auto p = in, end = in + inlen; p + alpnlen <= end; p += *p + 1) {
-    if (std::equal(alpn, alpn + alpnlen, p)) {
-      *out = p + 1;
-      *outlen = *p;
+  for (auto s = std::span{in, inlen}; s.size() >= HQ_ALPN_V1.size();
+       s = s.subspan(s[0] + 1)) {
+    if (std::ranges::equal(HQ_ALPN_V1, s.first(HQ_ALPN_V1.size()))) {
+      *out = &s[1];
+      *outlen = s[0];
       return SSL_TLSEXT_ERR_OK;
     }
   }
 
   if (!config.quiet) {
-    std::cerr << "Client did not present ALPN " << &alpn[1] << std::endl;
+    std::cerr << "Client did not present ALPN " << &HQ_ALPN_V1[1] << std::endl;
   }
 
   return SSL_TLSEXT_ERR_ALERT_FATAL;
